# -*- coding: utf-8 -*-

import os

from tqdm import tqdm
import requests


REPO_BASE_URL = (
    os.environ.get("TRANSLITERATE_HIN2ENG_DATA_URL")
    or "https://github.com/in-rolls/indicate/raw/master/indicate/data/model/"
)


def download_file(url, target):

    if "TRANSLITERATE_AUTH_TOKEN" in os.environ:
        auth_token = os.environ["TRANSLITERATE_AUTH_TOKEN"]
        headers = {"Authorization": "token {0!s}".format(auth_token)}
    else:
        headers = {}

    # Streaming, so we can iterate over the response.
    r = requests.get(url, stream=True, headers=headers)

    if r.status_code == 200:
        chunk_size = 64 * 1024
        # Total size in bytes.
        total_size = int(r.headers.get("content-length", 0)) / chunk_size

        total_size += 1

        with open(target, "wb") as f:
            for data in tqdm(r.iter_content(chunk_size), total=total_size, unit_scale=chunk_size / 1024, unit="KB"):
                f.write(data)
        return True
    else:
        print("ERROR: status_code={0:d}".format(r.status_code))
        return False



import tensorflow as tf


class Encoder(tf.keras.Model):
    def __init__(self, vocab_size, embedding_dim, enc_units, batch_sz):
        super(Encoder, self).__init__()
        self.batch_sz = batch_sz
        self.enc_units = enc_units
        self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_dim)

        # -------- LSTM layer in Encoder ------- ##
        self.lstm_layer = tf.keras.layers.LSTM(
            self.enc_units, return_sequences=True, return_state=True, recurrent_initializer="glorot_uniform"
        )

    def call(self, x, hidden):
        x = self.embedding(x)
        output, h, c = self.lstm_layer(x, initial_state=hidden)
        return output, h, c

    def initialize_hidden_state(self):
        return [tf.zeros((self.batch_sz, self.enc_units)), tf.zeros((self.batch_sz, self.enc_units))]



import sys
import argparse

from .hindi2english import HindiToEnglish

hindi2english = HindiToEnglish.transliterate


def main(argv=sys.argv[1:]):
    title = "Transliterate from Hindi to English"
    parser = argparse.ArgumentParser(description=title)
    parser.add_argument("--type", default=None, help="hin2eng")
    parser.add_argument("--input", default=None, help="Hindi text")
    args = parser.parse_args(argv)
    print(args)
    if not args.input or not args.type:
        return -1

    output = ""
    if args.type == "hin2eng":
        output = hindi2english(args.input)
    print(output)

    return 0


if __name__ == "__main__":
    sys.exit(main())



import tensorflow as tf
import tensorflow_addons as tfa

from .logging import get_logger

logger = get_logger()


class Decoder(tf.keras.Model):
    def __init__(
        self,
        vocab_size,
        embedding_dim,
        dec_units,
        batch_sz,
        max_length_input,
        max_length_output,
        attention_type="luong",
    ):
        super(Decoder, self).__init__()
        self.batch_sz = batch_sz
        self.dec_units = dec_units
        self.attention_type = attention_type
        self.max_length_input = max_length_input
        self.max_length_output = max_length_output

        # Embedding Layer
        self.embedding = tf.keras.layers.Embedding(vocab_size, embedding_dim)

        # Final Dense layer on which softmax will be applied
        self.fc = tf.keras.layers.Dense(vocab_size)

        # Define the fundamental cell for decoder recurrent structure
        self.decoder_rnn_cell = tf.keras.layers.LSTMCell(self.dec_units)

        # Sampler
        self.sampler = tfa.seq2seq.sampler.TrainingSampler()

        # Create attention mechanism with memory = None
        self.attention_mechanism = self.build_attention_mechanism(
            self.dec_units, None, self.batch_sz * [self.max_length_input], self.attention_type
        )

        # Wrap attention mechanism with the fundamental rnn cell of decoder
        self.rnn_cell = self.build_rnn_cell(batch_sz)

        # Define the decoder with respect to fundamental rnn cell
        self.decoder = tfa.seq2seq.BasicDecoder(self.rnn_cell, sampler=self.sampler, output_layer=self.fc)

    def build_rnn_cell(self, batch_sz):
        rnn_cell = tfa.seq2seq.AttentionWrapper(
            self.decoder_rnn_cell, self.attention_mechanism, attention_layer_size=self.dec_units
        )
        return rnn_cell

    def build_attention_mechanism(self, dec_units, memory, memory_sequence_length, attention_type="luong"):
        # ------------- #
        # typ: Which sort of attention (Bahdanau, Luong)
        # dec_units: final dimension of attention outputs
        # memory: encoder hidden states of shape (batch_size, max_length_input, enc_units)
        # memory_sequence_length: 1d array of shape (batch_size) with every element set to
        # max_length_input (for masking purpose)

        if attention_type == "bahdanau":
            return tfa.seq2seq.BahdanauAttention(
                units=dec_units, memory=memory, memory_sequence_length=memory_sequence_length
            )
        else:
            return tfa.seq2seq.LuongAttention(
                units=dec_units, memory=memory, memory_sequence_length=memory_sequence_length
            )

    def build_initial_state(self, batch_sz, encoder_state, Dtype):
        decoder_initial_state = self.rnn_cell.get_initial_state(batch_size=batch_sz, dtype=Dtype)
        decoder_initial_state = decoder_initial_state.clone(cell_state=encoder_state)
        return decoder_initial_state

    def call(self, inputs, initial_state):
        x = self.embedding(inputs)
        outputs, _, _ = self.decoder(
            x, initial_state=initial_state, sequence_length=self.batch_sz * [self.max_length_output - 1]
        )
        return outputs


# Generating text from sequence (mapping back)
def sequence_to_chars(tokenizer, sequence):
    word_index = tokenizer.word_index
    reverse_map = {val: key for key, val in word_index.items()}
    retext = ""
    for q in sequence:
        if q != 0:
            retext += reverse_map[q]
    return retext


def evaluate_sentence(sentence, units, input_lang_tokenizer, target_lang_tokenizer, encoder, decoder, max_length_input):
    logger.debug(f"Evaluating sentence {sentence}")
    # sentence = '^' + sentence.strip() + '$'

    inputs = [input_lang_tokenizer.word_index[i] for i in sentence]
    inputs = tf.keras.preprocessing.sequence.pad_sequences([inputs], maxlen=max_length_input, padding="post")
    inputs = tf.convert_to_tensor(inputs)
    inference_batch_size = inputs.shape[0]
    logger.debug(f"Inference batch size {inference_batch_size}")

    enc_start_state = [tf.zeros((inference_batch_size, units)), tf.zeros((inference_batch_size, units))]
    enc_out, enc_h, enc_c = encoder(inputs, enc_start_state)

    start_tokens = tf.fill([inference_batch_size], target_lang_tokenizer.word_index["^"])
    end_token = target_lang_tokenizer.word_index["$"]

    greedy_sampler = tfa.seq2seq.GreedyEmbeddingSampler()
    logger.debug("Initialized greedy sampler")

    # Instantiate BasicDecoder object
    decoder_instance = tfa.seq2seq.BasicDecoder(cell=decoder.rnn_cell, sampler=greedy_sampler, output_layer=decoder.fc)
    logger.debug("Initialized basic decoder")
    # Setup Memory in decoder stack
    decoder.attention_mechanism.setup_memory(enc_out)
    logger.debug("Attention memory setup done.")

    # set decoder_initial_state
    decoder_initial_state = decoder.build_initial_state(inference_batch_size, [enc_h, enc_c], tf.float32)
    logger.debug("Decoder initial state done.")

    # Since the BasicDecoder wraps around Decoder's rnn cell only, you have to ensure that the inputs to BasicDecoder
    # decoding step is output of embedding layer. tfa.seq2seq.GreedyEmbeddingSampler() takes care of this.
    # You only need to get the weights of embedding layer,
    # which can be done by decoder.embedding.variables[0] and pass this callabble to BasicDecoder's call() function

    decoder_embedding_matrix = decoder.embedding.variables[0]
    logger.debug("Assigned embedded weights")

    outputs, _, _ = decoder_instance(
        decoder_embedding_matrix, start_tokens=start_tokens, end_token=end_token, initial_state=decoder_initial_state
    )
    logger.debug("Got outputs from decoder instance.")
    return outputs.sample_id.numpy()


def transliterate(sentence, units, input_lang_tokenizer, target_lang_tokenizer, encoder, decoder, max_length_input):
    result = ""
    try:
        result = evaluate_sentence(
            sentence, units, input_lang_tokenizer, target_lang_tokenizer, encoder, decoder, max_length_input
        )
        result = sequence_to_chars(target_lang_tokenizer, result[0])
    except Exception:
        logger.error(f"Not able to transliterate {sentence}")
    return result.strip("$")



import signal
from .logging import get_logger
from .base import Base

import tensorflow as tf

from .encoder import Encoder
from .decoder import Decoder, transliterate

logger = get_logger()


class TimeoutException(Exception):  # Custom exception class
    def __init__(self, message, errors):
        super(TimeoutException, self).__init__(message)
        self.errors = errors


def timeout_handler(signum, frame):  # Custom signal handler
    raise TimeoutException("Timeout!")


# Change the behavior of SIGALRM
signal.signal(signal.SIGALRM, timeout_handler)


class HindiToEnglish(Base):
    MODELFN = "data/model/hindi_to_english/saved_weights/"
    INPUT_VOCAB = "data/model/hindi_to_english/hindi_tokens.json"
    TARGET_VOCAB = "data/model/hindi_to_english/english_tokens.json"

    embedding_dim = 256
    units = 1024
    BATCH_SIZE = 64
    BUFFER_SIZE = 120000

    max_length_input = 47
    max_length_output = 49
    START_TOKEN = "^"
    END_TOKEN = "$"

    weights_loaded = False

    input_lang_tokenizer = None
    target_lang_tokenizer = None
    encoder = None
    decoder = None

    @classmethod
    def transliterate(cls, input, latest=False):
        """
        Transliterate from Hindi to English.
        Args:
            input (str): Hindi text
        Returns:
            output (str): English text
        """

        if not cls.weights_loaded:
            cls.model_path, cls.input_vocab, cls.target_vocab = cls.load_model_data(latest)

            cls.input_lang_tokenizer = tf.keras.preprocessing.text.tokenizer_from_json(cls.input_vocab)
            cls.target_lang_tokenizer = tf.keras.preprocessing.text.tokenizer_from_json(cls.target_vocab)

            vocab_inp_size = len(cls.input_lang_tokenizer.word_index) + 1
            vocab_tar_size = len(cls.target_lang_tokenizer.word_index) + 1

            cls.encoder = Encoder(vocab_inp_size, cls.embedding_dim, cls.units, cls.BATCH_SIZE)
            cls.decoder = Decoder(
                vocab_tar_size,
                cls.embedding_dim,
                cls.units,
                cls.BATCH_SIZE,
                cls.max_length_input,
                cls.max_length_output,
                "luong",
            )
            logger.debug(f"checking weights for encoder {cls.encoder.embedding.variables}")
            logger.debug(f"checking weights for decoder {cls.decoder.embedding.variables}")

            # This is needed to load waits from checkpoint
            example_input_batch = tf.random.uniform(shape=[64, 47])
            sample_hidden = cls.encoder.initialize_hidden_state()
            sample_output, sample_h, sample_c = cls.encoder(example_input_batch, sample_hidden)
            sample_x = tf.random.uniform((cls.BATCH_SIZE, cls.max_length_output))
            cls.decoder.attention_mechanism.setup_memory(sample_output)
            initial_state = cls.decoder.build_initial_state(cls.BATCH_SIZE, [sample_h, sample_c], tf.float32)
            cls.decoder(sample_x, initial_state)

            logger.debug(f"checking weights for encoder after random input {cls.encoder.embedding.variables}")
            logger.debug(f"checking weights for decoder after random input {cls.decoder.embedding.variables}")

            logger.debug(f"Restoring model weights from {cls.model_path}")
            # restore weights from checkpoint
            cls.encoder.load_weights(f"{cls.model_path}/encoder")
            cls.decoder.load_weights(f"{cls.model_path}/decoder")
            logger.debug(f"encoder weights after restoring from checkpoint {cls.encoder.embedding.variables}")
            logger.debug(f"decoder weights after restoring from checkpoint {cls.decoder.embedding.variables}")

            cls.weights_loaded = True

        # Predictions
        # this is needed because sometimes BasicDecoder is going into loop
        # and not exiting for sentences like 'पति स्व.गंगा भगत'
        signal.alarm(10)
        target = ""
        try:
            target = transliterate(
                input,
                cls.units,
                cls.input_lang_tokenizer,
                cls.target_lang_tokenizer,
                cls.encoder,
                cls.decoder,
                cls.max_length_input,
            )
            logger.debug(f"Model predicted {target}")
        except TimeoutException as te:
            logger.error(f"Giving up on transliterating {input} due to {te}")
        except Exception as exe:
            logger.error(f"Not able to transliterate {input} due to {exe}")
        else:
            signal.alarm(0)

        return target



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import json

from pkg_resources import resource_filename
from .utils import download_file, REPO_BASE_URL
from .logging import get_logger

logger = get_logger()


class Base(object):
    MODELFN = None
    INPUT_VOCAB = None
    TARGET_VOCAB = None

    @classmethod
    def load_model_data(cls, latest=False):
        model_path = None
        input_vocab = None
        target_vocab = None
        failed = False

        if cls.MODELFN:
            model_fn = resource_filename(__name__, cls.MODELFN)
            path = os.path.dirname(model_fn)
            if not os.path.exists(path):
                os.makedirs(path)
            if not os.path.exists(model_fn) or latest:
                logger.debug("Downloading model data from the server ({0!s})...".format(model_fn))
                if not download_file(REPO_BASE_URL + cls.MODELFN, model_fn):
                    logger.error("ERROR: Cannot download model data file")
                    failed = True
            else:
                logger.debug("Using model data from {0!s}...".format(model_fn))
            model_path = path

        if cls.INPUT_VOCAB:
            input_vocab_file = resource_filename(__name__, cls.INPUT_VOCAB)
            path = os.path.dirname(input_vocab_file)
            if not os.path.exists(path):
                os.makedirs(path)
            if not os.path.exists(input_vocab_file) or latest:
                logger.debug("Downloading model data from the server ({0!s})...".format(input_vocab_file))
                if not download_file(REPO_BASE_URL + cls.INPUT_VOCAB, input_vocab_file):
                    logger.error("ERROR: Cannot download input vocabulary file")
                    failed = True
            else:
                logger.debug("Using input vocab data from {0!s}...".format(input_vocab_file))

        if cls.TARGET_VOCAB:
            output_vocab_file = resource_filename(__name__, cls.TARGET_VOCAB)
            path = os.path.dirname(output_vocab_file)
            if not os.path.exists(path):
                os.makedirs(path)
            if not os.path.exists(output_vocab_file) or latest:
                logger.debug("Downloading model data from the server ({0!s})...".format(output_vocab_file))
                if not download_file(REPO_BASE_URL + cls.TARGET_VOCAB, output_vocab_file):
                    logger.error("ERROR: Cannot download target vocabulary file")
                    failed = True
            else:
                logger.debug("Using output vocab data from {0!s}...".format(output_vocab_file))

        if not failed:
            with open(input_vocab_file, "r") as fh:
                input_vocab = json.load(fh)
            with open(output_vocab_file, "r") as fh:
                target_vocab = json.load(fh)

        return model_path, input_vocab, target_vocab



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.test import test as TestCommand


# Get the long description from the README file
with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        print("TODO: PostDevelopCommand")
        develop.run(self)


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        print("TODO: PostInstallCommand")
        install.run(self)


class Tox(TestCommand):
    user_options = [("tox-args=", "a", "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        import shlex

        args = self.tox_args
        if args:
            args = shlex.split(self.tox_args)
        tox.cmdline(args=args)


setup(
    name="indicate",
    version="0.0.8",
    description="Transliterations to/from Indian languages",
    long_description=long_description,
    # The project's main homepage.
    url="https://github.com/in-rolls/indicate",
    # Author details
    author="Rajashekar Chintalapati, Gaurav Sood",
    author_email="rajshekar.ch@gmail.com, gsood07@gmail.com",
    # Choose your license
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    # What does your project relate to?
    keywords="transliterate indic hindi english",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["data", "docs", "tests", "scripts"]),
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["tqdm", "tensorflow", "tensorflow_addons"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": ["check-manifest"],
        "test": ["coverage"],
    },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        "indicate": [
            "data/model/hindi_to_english/saved_weights/checkpoint",
            "data/model/hindi_to_english/saved_weights/encoder.*",
            "data/model/hindi_to_english/saved_weights/decoder.*",
            "data/model/hindi_to_english/*.json",
            "notebooks/*.ipynb",
        ],
    },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "console_scripts": ["hindi2english=indicate.transliterate:main"],
    },
    cmdclass={
        "develop": PostDevelopCommand,
        "install": PostInstallCommand,
        "test": Tox,
    },
    tests_require=["tox"],
)
